/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package com.qiniu.android.dns

import com.qiniu.android.dns.http.DomainNotOwn
import com.qiniu.android.dns.local.Hosts
import com.qiniu.android.dns.util.LruCache
import java.io.IOException
import java.net.InetAddress
import java.net.UnknownHostException
import java.util.Date
import java.util.TimeZone
import java.util.concurrent.atomic.AtomicInteger
import kotlin.concurrent.Volatile

/** DNS解析管理类，可以重复使用 */
class DnsManager
@JvmOverloads
constructor(info: NetworkInfo?, resolvers: Array<IResolver>, sorter: RecordSorter? = null) {
    private val resolvers: Array<IResolver>
    private val cache: LruCache<String?, Array<Record?>?>
    private val hosts = Hosts()
    private val sorter: RecordSorter
    private var queryErrorHandler: QueryErrorHandler? = null

    @Volatile private var info: NetworkInfo

    @Volatile private var index = 0
    /**
     * @param info 当前的网络信息，从Android context中获取
     * @param resolvers 具体的dns 解析示例，可以是local或者httpdns
     * @param sorter 外部接口，对获取的IP数组进行排序
     */
    /**
     * @param info 当前的网络信息，从Android context中获取
     * @param resolvers 具体的dns 解析示例，可以是local或者httpdns
     */
    init {
        this.info = info ?: NetworkInfo.normal
        this.resolvers = resolvers.clone()
        cache = LruCache()
        this.sorter = sorter ?: DummySorter()
    }

    /**
     * 查询域名
     *
     * @param domain 域名参数
     * @return ip 记录 列表
     * @throws IOException 网络异常或者无法解析抛出异常
     */
    @Throws(IOException::class)
    fun queryRecords(domain: String?): Array<out Record?>? {
        return queryRecords(Domain(domain))
    }

    @Throws(IOException::class)
    fun queryRecords(domain: Domain?): Array<out Record?>? {
        if (domain == null) {
            throw IOException("null domain")
        }
        if (domain.domain == null || domain.domain.trim { it <= ' ' }.isEmpty()) {
            throw IOException("empty domain " + domain.domain)
        }
        if (validIP(domain.domain)) {
            val record =
                Record(
                    domain.domain,
                    Record.TYPE_A,
                    Record.TTL_Forever,
                    Date().time,
                    Record.Source.Unknown)
            return arrayOf(record)
        }
        val records = queryRecordInternal(domain)
        return sorter.sort(records)
    }

    @Throws(IOException::class)
    private fun queryRecordInternal(domain: Domain): Array<Record?>? {
        if (domain.hostsFirst) {
            var ret = hosts.query(domain, info)
            ret = filterInvalidRecords(ret)
            if (!ret.isNullOrEmpty()) {
                return ret
            }
        }
        synchronized(cache) {
            if (info == NetworkInfo.normal && Network.isNetworkChanged) {
                cache.clear()
                synchronized(resolvers) { index = 0 }
            } else {
                var records = cache[domain.domain]
                records = filterInvalidRecords(records)
                if (records != null && records!!.isNotEmpty()) {
                    return records
                }
            }
        }
        var records: Array<Record?>? = null
        var lastE: IOException? = null
        val firstOk = index
        for (i in resolvers.indices) {
            val pos = (firstOk + i) % resolvers.size
            val before = info
            val ip = Network.ip
            try {
                records = resolvers[pos].resolve(domain, info)
            } catch (e: DomainNotOwn) {
                continue
            } catch (e: IOException) {
                lastE = e
                if (queryErrorHandler != null) {
                    queryErrorHandler!!.queryError(e, domain.domain)
                }
            } catch (e2: Exception) {
                lastE = IOException(e2)
                if (queryErrorHandler != null) {
                    queryErrorHandler!!.queryError(e2, domain.domain)
                }
            }
            val ip2 = Network.ip
            if (info == before && (records == null || records!!.isEmpty()) && ip == ip2) {
                synchronized(resolvers) {
                    if (index == firstOk) {
                        index++
                        if (index == resolvers.size) {
                            index = 0
                        }
                    }
                }
            } else {
                break
            }
        }
        if (records == null || records!!.isEmpty()) {
            if (!domain.hostsFirst) {
                var rs = hosts.query(domain, info)
                rs = filterInvalidRecords(rs)
                if (!rs.isNullOrEmpty()) {
                    return rs
                }
            }
            if (lastE != null) {
                throw lastE
            }
            val e: IOException = UnknownHostException(domain.domain)
            if (queryErrorHandler != null) {
                queryErrorHandler!!.queryError(e, domain.domain)
            }
            throw e
        }
        records = trimCname(records!!)
        if (records!!.isEmpty()) {
            throw UnknownHostException("no A/AAAA records")
        }
        synchronized(cache) { cache.put(domain.domain, records) }
        return records
    }

    @Throws(IOException::class)
    fun queryInetAdress(domain: Domain?): Array<InetAddress?>? {
        val records = queryRecords(domain)
        val ips = records2Ip(records)
        if (ips.isNullOrEmpty()) {
            return null
        }
        val addresses = arrayOfNulls<InetAddress>(ips.size)
        for (i in ips.indices) {
            addresses[i] = InetAddress.getByName(ips[i])
        }
        return addresses
    }

    /**
     * 当网络发生变化时，通知当前的网络信息
     *
     * @param info 网络信息
     */
    fun onNetworkChange(info: NetworkInfo?) {
        clearCache()
        this.info = info ?: NetworkInfo.normal
        synchronized(resolvers) { index = 0 }
    }

    private fun clearCache() {
        synchronized(cache) { cache.clear() }
    }

    /**
     * 插入指定运营商的hosts配置
     *
     * @param domain 域名
     * @param record record 记录, source 配置为 @[Record.Source.Custom]
     * @param provider 运营商，见 NetworkInfo
     * @return 当前的DnsManager，便于链式调用
     */
    fun putHosts(domain: String?, record: Record, provider: Int): DnsManager {
        val recordNew =
            Record(
                record.value,
                record.type,
                record.ttl,
                record.timeStamp,
                Record.Source.Custom,
                record.server)
        hosts.put(domain!!, Hosts.Value(recordNew, provider))
        return this
    }

    /**
     * 插入指定运营商的hosts配置
     *
     * @param domain 域名
     * @param type ip 类别，[Record.TYPE_A] / [Record.TYPE_AAAA]
     * @param ip ip
     * @param provider 运营商，见 NetworkInfo
     * @return 当前的DnsManager，便于链式调用
     */
    fun putHosts(domain: String?, type: Int, ip: String, provider: Int): DnsManager {
        putHosts(
            domain,
            Record(ip, type, Record.TTL_Forever, Date().time / 1000, Record.Source.Custom),
            provider)
        return this
    }

    /**
     * 插入指定运营商的hosts配置
     *
     * @param domain 域名
     * @param type ip 类别，[Record.TYPE_A] / [Record.TYPE_AAAA]
     * @param ip ip
     * @return 当前的DnsManager，便于链式调用
     */
    fun putHosts(domain: String?, type: Int, ip: String): DnsManager {
        putHosts(domain, type, ip, NetworkInfo.ISP_GENERAL)
        return this
    }

    /**
     * 插入指定运营商的hosts配置
     *
     * @param domain 域名
     * @param ipv4 ipv4 ip
     * @return 当前的DnsManager，便于链式调用
     */
    fun putHosts(domain: String?, ipv4: String): DnsManager {
        putHosts(domain, Record.TYPE_A, ipv4)
        return this
    }

    interface QueryErrorHandler {
        fun queryError(e: Exception?, host: String?)
    }

    private class DummySorter : RecordSorter {
        private val pos = AtomicInteger()

        override fun sort(records: Array<Record?>?): Array<Record?>? {
            return records
        }
    }

    companion object {
        private fun trimCname(records: Array<Record?>): Array<Record?> {
            val a = ArrayList<Record?>(records.size)
            for (r in records) {
                if (r != null && (r.isA || r.isAAAA)) {
                    a.add(r)
                }
            }
            return a.toTypedArray<Record?>()
        }

        private fun records2Ip(records: Array<out Record?>?): Array<String>? {
            if (records == null || records.isEmpty()) {
                return null
            }
            val a = ArrayList<String>(records.size)
            for (r in records) {
                r?.let { a.add(r.value) }
            }
            return if (a.size == 0) {
                null
            } else a.toTypedArray<String>()
        }

        private fun filterInvalidRecords(records: Array<Record?>?): Array<Record?>? {
            if (records.isNullOrEmpty()) {
                return null
            }
            val ret = ArrayList<Record?>(records.size)
            for (r in records) {
                if (r?.value != null && r.value.isNotEmpty() && !r.isExpired) {
                    ret.add(r)
                }
            }
            return if (ret.size == 0) {
                null
            } else ret.toTypedArray<Record?>()
        }

        fun validIP(ip: String?): Boolean {
            if (ip == null || ip.length < 7 || ip.length > 15) return false
            return if (ip.contains("-")) false
            else
                try {
                    var x = 0
                    var y = ip.indexOf('.')
                    if (y != -1 && ip.substring(x, y).toInt() > 255) return false
                    x = ip.indexOf('.', ++y)
                    if (x != -1 && ip.substring(y, x).toInt() > 255) return false
                    y = ip.indexOf('.', ++x)
                    !(y != -1 &&
                        ip.substring(x, y).toInt() > 255 &&
                        ip.substring(++y, ip.length - 1).toInt() > 255 &&
                        ip[ip.length - 1] != '.')
                } catch (e: NumberFormatException) {
                    false
                }
        }

        // 简单通过时区判断是否在大陆
        fun needHttpDns(): Boolean {
            return try {
                val zone = TimeZone.getDefault()
                val id = zone.id
                "Asia/Shanghai" == id ||
                    "Asia/Chongqing" == id ||
                    "Asia/Harbin" == id ||
                    "Asia/Urumqi" == id
            } catch (e: Exception) {
                false
            }
        }
    }
}
