/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package com.qiniu.android.dns

import java.util.Date
import java.util.Locale

/** Created by bailong on 15/6/12. */
class Record {
    /** 具体的值，A 记录时为IP，CName时为指向的域名 */
    val value: String

    /** 记录类型，A或者CName */
    @JvmField val type: Int

    /** TTL dns结果缓存时间 */
    val ttl: Int

    /** 时间戳，用来判断超时，单位：秒 */
    val timeStamp: Long

    /** 记录来源， httpDns 或者 System [Source] */
    val source: Int
    val server: String?

    constructor(value: String, type: Int, ttl: Int) {
        this.value = value
        this.type = type
        this.ttl = ttl
        timeStamp = Date().time / 1000
        source = Source.Unknown
        server = null
    }

    constructor(value: String, type: Int, ttl: Int, timeStamp: Long, source: Int) {
        this.value = value
        this.type = type
        this.ttl = ttl.coerceAtLeast(TTL_MIN_SECONDS)
        this.timeStamp = timeStamp
        this.source = source
        server = null
    }

    constructor(value: String, type: Int, ttl: Int, timeStamp: Long, source: Int, server: String?) {
        this.value = value
        this.type = type
        this.ttl = ttl.coerceAtLeast(TTL_MIN_SECONDS)
        this.timeStamp = timeStamp
        this.source = source
        this.server = server
    }

    override fun equals(other: Any?): Boolean {
        if (this === other) {
            return true
        }
        return if (other !is Record) {
            false
        } else
            value == other.value &&
                type == other.type &&
                ttl == other.ttl &&
                timeStamp == other.timeStamp
    }

    val isA: Boolean
        get() = type == TYPE_A

    val isAAAA: Boolean
        get() = type == TYPE_AAAA

    val isCname: Boolean
        get() = type == TYPE_CNAME

    val isExpired: Boolean
        get() = isExpired(System.currentTimeMillis() / 1000)

    private fun isExpired(time: Long): Boolean {
        return if (ttl == TTL_Forever) {
            false
        } else timeStamp + ttl < time
    }

    override fun toString(): String {
        return String.format(
            Locale.getDefault(),
            "{type:%s, value:%s, source:%s, server:%s, timestamp:%d, ttl:%d}",
            type,
            value,
            source,
            server,
            timeStamp,
            ttl)
    }

    object Source {
        const val Unknown = 0
        const val Custom = 1
        const val DnspodEnterprise = 2
        const val System = 3
        const val Udp = 4
        const val Doh = 5
    }

    companion object {
        const val TTL_MIN_SECONDS = 600
        const val TTL_Forever = -1

        /** A 记录 类型 */
        const val TYPE_A = 1

        /** AAAA 记录 类型 */
        const val TYPE_AAAA = 28

        /** CName 类型 */
        const val TYPE_CNAME = 5

        /** TXT 记录 类型 */
        const val TYPE_TXT = 16
    }
}
