/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package com.qiniu.android.dns.dns

import com.qiniu.android.dns.Record
import java.io.IOException
import java.net.IDN
import java.util.Arrays
import java.util.Date
import java.util.Locale

internal class DnsResponse(
    server: String,
    source: Int,
    request: DnsRequest,
    recordData: ByteArray?
) : DnsMessage() {
    private val timestamp: Long
    private val source: Int
    private val server: String
    private val request: DnsRequest
    private val recordData: ByteArray
    private var aA = 0
    private var rCode = 0
    var answerArray: List<Record>? = null
    private var authorityArray: List<Record>? = null
    private var additionalArray: List<Record>? = null

    init {
        if (recordData == null || recordData.isEmpty()) {
            throw IOException("response data is empty")
        }
        this.server = server
        this.source = source
        this.request = request
        this.recordData = recordData
        timestamp = Date().time / 1000
        parse()
    }

    @Throws(IOException::class)
    private fun parse() {
        if (recordData.size < 12) {
            throw IOException("response data too small")
        }

        // Header
        parseHeader()

        // Question
        var index = parseQuestion()

        // Answer
        val answer = RecordResource("answer", readRecordDataInt16(6).toInt(), index)
        parseResourceRecord(answer)
        answerArray = answer.records
        index += answer.length

        // Authority
        val authority = RecordResource("authority", readRecordDataInt16(8).toInt(), index)
        parseResourceRecord(authority)
        authorityArray = authority.records
        index += authority.length

        // Additional
        val additional = RecordResource("additional", readRecordDataInt16(10).toInt(), index)
        parseResourceRecord(additional)
        additionalArray = additional.records
    }

    @Throws(IOException::class)
    private fun parseHeader() {
        messageId = readRecordDataInt16(0)

        // question id 不匹配
        if (messageId != request.messageId) {
            throw IOException("question id error")
        }

        // |00|01|02|03|04|05|06|07|
        // |QR|  OPCODE   |AA|TC|RD|
        val field0 = readRecordDataInt8(2)
        val qr = readRecordDataInt8(2) and 0x80
        // 非 dns 响应数据
        if (qr == 0) {
            throw IOException("not a response data")
        }
        opCode = field0 shr 3 and 0x0f // opCode is 4 bit
        aA = field0 shr 2 and 0x01
        rD = field0 and 0x01

        // |00|01|02|03|04|05|06|07|
        // |RA|r1|r2|r3| RCODE     |
        val field1 = readRecordDataInt8(3)
        rA = field1 shr 7 and 0x01
        rCode = field1 and 0x0F
    }

    @Throws(IOException::class)
    private fun parseQuestion(): Int {
        var index = 12
        var qdCount = readRecordDataInt16(4).toInt()
        while (qdCount > 0) {
            val recordName = getNameFrom(index) ?: throw IOException("read Question error")
            index += recordName.skipLength + 4
            qdCount--
        }
        return index
    }

    @Throws(IOException::class)
    private fun parseResourceRecord(resource: RecordResource) {
        var index = resource.from
        var count = resource.count
        while (count > 0) {
            val recordName =
                getNameFrom(index) ?: throw IOException("read " + resource.name + " error")
            index += recordName.skipLength
            val type = readRecordDataInt16(index).toInt()
            index += 2
            val clazz = readRecordDataInt16(index).toInt()
            index += 2
            val ttl = readRecordDataInt32(index)
            index += 4
            val rdLength = readRecordDataInt16(index).toInt()
            index += 2
            val value = readData(type, index, rdLength)
            if (clazz == 0x01 && (type == Record.TYPE_CNAME || type == request.recordType)) {
                val record = Record(value!!, type, ttl, timestamp, source, server)
                resource.addRecord(record)
            }
            index += rdLength
            count--
        }
        resource.length = index - resource.from
    }

    @Throws(IOException::class)
    private fun getNameFrom(from: Int): RecordName? {
        var partLength: Int
        var index = from
        val name = StringBuilder()
        val recordName = RecordName()
        var maxLoop = 128
        do {
            partLength = readRecordDataInt8(index)
            if (partLength and 0xc0 == 0xc0) {
                // name pointer
                if (recordName.skipLength < 1) {
                    recordName.skipLength = index + 2 - from
                }
                index = partLength and 0x3f shl 8 or readRecordDataInt8(index + 1)
                continue
            } else if (partLength and 0xc0 > 0) {
                return null
            } else {
                index++
            }
            if (partLength > 0) {
                if (name.isNotEmpty()) {
                    name.append(".")
                }
                val nameData = Arrays.copyOfRange(recordData, index, index + partLength)
                name.append(IDN.toUnicode(String(nameData)))
                index += partLength
            }
        } while (partLength > 0 && --maxLoop > 0)
        recordName.name = name.toString()
        if (recordName.skipLength < 1) {
            recordName.skipLength = index - from
        }
        return recordName
    }

    @Throws(IOException::class)
    private fun readData(recordType: Int, from: Int, length: Int): String? {
        var dataString: String? = null
        when (recordType) {
            Record.TYPE_A -> {
                if (length == 4) {
                    val builder = StringBuilder()
                    builder.append(readRecordDataInt8(from))
                    for (i in 1..3) {
                        builder.append(".")
                        builder.append(readRecordDataInt8(from + i))
                    }
                    dataString = builder.toString()
                }
            }
            Record.TYPE_AAAA -> {
                if (length == 16) {
                    val builder = StringBuilder()
                    var i = 0
                    while (i < 16) {
                        builder.append(if (i > 0) ":" else "")
                        Integer.toHexString(readRecordDataInt16(from + i).toInt())
                        i += 2
                    }
                    dataString = builder.toString()
                }
            }
            Record.TYPE_CNAME -> {
                if (length > 1) {
                    val name = getNameFrom(from)!!
                    dataString = name.name
                }
            }
            Record.TYPE_TXT -> {
                if (length > 0 && from + length < recordData.size) {
                    val data = Arrays.copyOfRange(recordData, from, from + length)
                    val dataValue = String(data)
                    dataString = IDN.toUnicode(dataValue)
                }
            }
            else -> {}
        }
        return dataString
    }

    @Throws(IOException::class)
    private fun readRecordDataInt8(from: Int): Int {
        if (from >= recordData.size) {
            throw IOException("read response data out of range")
        }
        return recordData[from].toInt() and 0xFF
    }

    @Throws(IOException::class)
    private fun readRecordDataInt16(from: Int): Short {
        if (from + 1 >= recordData.size) {
            throw IOException("read response data out of range")
        }
        val b0 = recordData[from].toInt() and 0xFF shl 8
        val b1 = recordData[from + 1].toInt() and 0xFF
        return (b0 + b1).toShort()
    }

    @Throws(IOException::class)
    private fun readRecordDataInt32(from: Int): Int {
        if (from + 3 >= recordData.size) {
            throw IOException("read response data out of range")
        }
        val b0 = recordData[from].toInt() and 0xFF shl 24
        val b1 = recordData[from + 1].toInt() and 0xFF shl 16
        val b2 = recordData[from + 2].toInt() and 0xFF shl 8
        val b3 = recordData[from + 3].toInt() and 0xFF
        return b0 + b1 + b2 + b3
    }

    override fun toString(): String {
        return String.format(
            Locale.getDefault(),
            "{messageId:%d, rd:%d, ra:%d, aa:%d, rCode:%d, server:%s, request:%s, answerArray:%s, authorityArray:%s, additionalArray:%s}",
            messageId,
            rD,
            rA,
            aA,
            rCode,
            server,
            request,
            answerArray,
            authorityArray,
            additionalArray)
    }

    class RecordResource(val name: String, val count: Int, val from: Int) {
        val records: MutableList<Record>
        var length = 0

        init {
            records = ArrayList()
        }

        fun addRecord(record: Record?) {
            if (record != null) {
                records.add(record)
            }
        }
    }

    private class RecordName {
        var skipLength = 0
        var name: String? = null
    }
}
