/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package com.qiniu.android.dns.local

import android.annotation.SuppressLint
import android.content.Context
import android.net.ConnectivityManager
import android.net.ConnectivityManager.NetworkCallback
import android.net.LinkProperties
import android.net.Network
import android.net.NetworkRequest
import com.qiniu.android.dns.Domain
import com.qiniu.android.dns.IResolver
import com.qiniu.android.dns.NetworkInfo
import com.qiniu.android.dns.Record
import com.qiniu.android.dns.dns.DnsUdpResolver
import java.io.IOException
import java.io.InputStreamReader
import java.io.LineNumberReader
import java.net.InetAddress
import java.util.logging.Level
import java.util.logging.Logger

/** reference github/rtreffer/minidns. */
object AndroidDnsServer {
    val byCommand: List<InetAddress>?
        //    15ms
        get() {
            try {
                val process = Runtime.getRuntime().exec("getprop")
                val inputStream = process.inputStream
                val lnr = LineNumberReader(InputStreamReader(inputStream))
                var line: String
                val servers = ArrayList<InetAddress>(5)
                while (lnr.readLine().also { line = it } != null) {
                    val split = line.indexOf("]: [")
                    if (split <= 1 || line.length - 1 <= split + 4) {
                        continue
                    }
                    val property = line.substring(1, split)
                    var value = line.substring(split + 4, line.length - 1)
                    if (property.endsWith(".dns") ||
                        property.endsWith(".dns1") ||
                        property.endsWith(".dns2") ||
                        property.endsWith(".dns3") ||
                        property.endsWith(".dns4")) {

                        // normalize the address
                        val ip = InetAddress.getByName(value) ?: continue
                        value = ip.hostAddress
                        if (value.isNullOrEmpty()) continue
                        servers.add(ip)
                    }
                }
                if (servers.size > 0) {
                    return servers
                }
            } catch (e: IOException) {
                Logger.getLogger("AndroidDnsServer")
                    .log(Level.WARNING, "Exception in findDNSByExec", e)
            }
            return null
        }

    val byReflection: List<InetAddress>?
        // 1ms
        get() {
            try {
                @SuppressLint("PrivateApi")
                val SystemProperties = Class.forName("android.os.SystemProperties")
                val method = SystemProperties.getMethod("get", String::class.java)
                val servers = ArrayList<InetAddress>(5)
                for (propKey in arrayOf("net.dns1", "net.dns2", "net.dns3", "net.dns4")) {
                    var value = method.invoke(null, propKey) as String
                    if (value.isEmpty()) continue
                    val ip = InetAddress.getByName(value) ?: continue
                    value = ip.hostAddress
                    if (value.isNullOrEmpty()) continue
                    if (servers.contains(ip)) continue
                    servers.add(ip)
                }
                if (servers.size > 0) {
                    return servers
                }
            } catch (e: Exception) {
                // we might trigger some problems this way
                Logger.getLogger("AndroidDnsServer")
                    .log(Level.WARNING, "Exception in findDNSByReflection", e)
            }
            return null
        }

    fun defaultResolver(context: Context): IResolver {
        //        the system dns ip would change after network changed.
        return AndroidResolver(context)
    }

    internal class AndroidResolver(context: Context) : IResolver {
        private val dnsServers: MutableList<InetAddress> = ArrayList()
        private var networkCallback = false

        init {
            var addresses = byReflection
            if (addresses == null) {
                addresses = byCommand
            }
            if (addresses == null) {
                /// Android 8 , net.dns* was disabled, query dns servers must use network callback
                /// @see https://developer.android.com/about/versions/oreo/android-8.0-changes.html
                val connectivityManager =
                    context.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
                val builder = NetworkRequest.Builder()
                try {
                    connectivityManager.registerNetworkCallback(
                        builder.build(),
                        object : NetworkCallback() {
                            override fun onLinkPropertiesChanged(
                                network: Network,
                                linkProperties: LinkProperties
                            ) {
                                dnsServers.addAll(linkProperties.dnsServers)
                                networkCallback = true
                            }
                        })
                } catch (e: Exception) {
                    e.printStackTrace()
                }
            } else {
                dnsServers.addAll(addresses)
            }
        }

        @Throws(IOException::class)
        override fun resolve(domain: Domain?, info: NetworkInfo?): Array<Record?>? {
            if (dnsServers.isEmpty() && !networkCallback) {
                /// Network callback may delay, wait for 1000 ms
                try {
                    Thread.sleep(1000)
                } catch (e: InterruptedException) {
                    e.printStackTrace()
                }
            }
            if (dnsServers.isEmpty()) {
                throw IOException("cant get local dns server")
            }
            val dnsServer = dnsServers[0]
            val resolver: IResolver = HijackingDetectWrapper(DnsUdpResolver(dnsServer.hostName))
            val records = resolver.resolve(domain, info)
            domain?.let {
                if (domain.hasCname) {
                    var cname = false
                    records?.let {
                        for (r in records) {
                            if (r?.isCname == true) {
                                cname = true
                                break
                            }
                        }
                    }

                    if (!cname) {
                        throw DnshijackingException(domain.domain, dnsServer.hostAddress)
                    }
                }
                if (domain.maxTtl != 0) {
                    records?.let {
                        for (r in records) {
                            if (r?.isCname == false) {
                                if (r.ttl > domain.maxTtl) {
                                    throw DnshijackingException(
                                        domain.domain, dnsServer.hostAddress, r.ttl)
                                }
                            }
                        }
                    }
                }
            }
            return records
        }
    }
}
