/*
 * Copyright (c) 田梓萱[小草林] 2021-2024.
 * All Rights Reserved.
 * All codes are protected by China's regulations on the protection of computer software, and infringement must be investigated.
 * 版权所有 (c) 田梓萱[小草林] 2021-2024.
 * 所有代码均受中国《计算机软件保护条例》保护，侵权必究.
 */
package com.qiniu.android.dns.util

import java.io.IOException

/**
 * Converts hexadecimal Strings. This class is thread-safe.
 *
 * @since 1.1
 * @version $Id: Hex.java 1619948 2014-08-22 22:53:55Z ggregory $
 */
object Hex {
    /** Used to build output as Hex */
    private val DIGITS_LOWER =
        charArrayOf('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f')

    /** Used to build output as Hex */
    private val DIGITS_UPPER =
        charArrayOf('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F')

    /**
     * Converts an array of characters representing hexadecimal values into an array of bytes of
     * those same values. The returned array will be half the length of the passed array, as it
     * takes two characters to represent any given byte. An exception is thrown if the passed char
     * array has an odd number of elements.
     *
     * @param data An array of characters containing hexadecimal digits
     * @return A byte array containing binary data decoded from the supplied char array.
     * @throws HexDecodeException Thrown if an odd number or illegal of characters is supplied
     */
    @Throws(HexDecodeException::class)
    fun decodeHex(data: CharArray): ByteArray {
        val len = data.size
        if (len and 0x01 != 0) {
            throw HexDecodeException("Odd number of characters.")
        }
        val out = ByteArray(len shr 1)

        // two characters form the hex value.
        var i = 0
        var j = 0
        while (j < len) {
            var f = toDigit(data[j], j) shl 4
            j++
            f = f or toDigit(data[j], j)
            j++
            out[i] = (f and 0xFF).toByte()
            i++
        }
        return out
    }
    /**
     * Converts an array of bytes into an array of characters representing the hexadecimal values of
     * each byte in order. The returned array will be double the length of the passed array, as it
     * takes two characters to represent any given byte.
     *
     * @param data a byte[] to convert to Hex characters
     * @param toLowerCase `true` converts to lowercase, `false` to uppercase
     * @return A char[] containing hexadecimal characters
     * @since 1.4
     */
    /**
     * Converts an array of bytes into an array of characters representing the hexadecimal values of
     * each byte in order. The returned array will be double the length of the passed array, as it
     * takes two characters to represent any given byte.
     *
     * @param data a byte[] to convert to Hex characters
     * @return A char[] containing hexadecimal characters
     */
    @JvmOverloads
    fun encodeHex(data: ByteArray, toLowerCase: Boolean = true): CharArray {
        return encodeHex(data, if (toLowerCase) DIGITS_LOWER else DIGITS_UPPER)
    }

    /**
     * Converts an array of bytes into an array of characters representing the hexadecimal values of
     * each byte in order. The returned array will be double the length of the passed array, as it
     * takes two characters to represent any given byte.
     *
     * @param data a byte[] to convert to Hex characters
     * @param toDigits the output alphabet
     * @return A char[] containing hexadecimal characters
     * @since 1.4
     */
    private fun encodeHex(data: ByteArray, toDigits: CharArray): CharArray {
        val l = data.size
        val out = CharArray(l shl 1)
        // two characters form the hex value.
        var i = 0
        var j = 0
        while (i < l) {
            out[j++] = toDigits[0xF0 and data[i].toInt() ushr 4]
            out[j++] = toDigits[0x0F and data[i].toInt()]
            i++
        }
        return out
    }

    /**
     * Converts an array of bytes into a String representing the hexadecimal values of each byte in
     * order. The returned String will be double the length of the passed array, as it takes two
     * characters to represent any given byte.
     *
     * @param data a byte[] to convert to Hex characters
     * @return A String containing hexadecimal characters
     * @since 1.4
     */
    fun encodeHexString(data: ByteArray): String {
        return String(encodeHex(data))
    }

    /**
     * Converts a hexadecimal character to an integer.
     *
     * @param ch A character to convert to an integer digit
     * @param index The index of the character in the source
     * @return An integer
     * @throws HexDecodeException Thrown if ch is an illegal hex character
     */
    @Throws(HexDecodeException::class)
    private fun toDigit(ch: Char, index: Int): Int {
        val digit = ch.digitToIntOrNull(16) ?: -1
        if (digit == -1) {
            throw HexDecodeException("Illegal hexadecimal character $ch at index $index")
        }
        return digit
    }

    class HexDecodeException(msg: String?) : IOException(msg)
}
